/* more.c */

/* A program to display the contents of a text file, one screenful at
   a time */

/* Copyright (C) 1994-2002 Jim Hall <jhall@freedos.org> */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    See the file 'COPYING' for details.
*/


#include <stdio.h>
#include <stdlib.h>			/* for _splitpath, _makepath */

#include <dir.h>			/* for findfirst, findnext */
#include <dos.h>			/* for findfirst, findnext */
#include <bios.h>			/* for _bios_keybrd - see keypress() */
#include <conio.h>			/* for getch - see keypress() */

#include "catgets.h"			/* Cats message library */


/* Symbolic constants */

#define LINES 25			/* lines per screen */
#define COLS  80			/* columns per screen */

#define CTRL_C 3			/* control-c */


/* Function prototypes */

int more (FILE *pfile, const char *descr, const char *prompt);
unsigned keypress (void);
void usage (nl_catd cat);


/* Main program */

int
main (int argc, char **argv)
{
  char *s, *prompt;
  char drive[_MAX_DRIVE];		/* for _splitpath, _makepath */
  char dir[_MAX_DIR];			/* for _splitpath, _makepath */
  char file[_MAX_FNAME];		/* for _splitpath, _makepath */
  char ext[_MAX_EXT];			/* for _splitpath, _makepath */
  char fullpath[_MAX_PATH];		/* the full path after _makepath */

  int i;
  int next;				/* return value of more() function */
  int done;				/* flag for findfirst, findnext */
  int exitval;				/* exit value */

  FILE *pfile;				/* file handle */
  struct ffblk ffblk;			/* findfirst, findnext block */
  nl_catd cat;				/* message catalog */

  /* Open the message catalog */

  cat = catopen ("more", 0);
  prompt = catgets (cat, 2, 0, "More");

  /* Evaluate the files on the command line */

  /* Yes, it would be better and more standard to use getopt() to
     process the list of command line args, but since MORE does not
     take any options, we can simply the scanning and just look for
     '/' in the command line. */

  if (argc == 1)
    {
      /* Process the standard input.  Return from more() is not used. */

      s = catgets (cat, 2, 1, "<STDIN>");
      more (stdin, s, prompt);
      exitval = 0;
    }

  for (i = 1; i < argc; i++)
    {
      /* Check if we are asking for help */

      if (argv[i][0] == '/')
	{
	  /* This is a flag character */

	  /* Add any code here that deals with special flag chars */

	  switch (argv[i][1])
	    {
	    case '?':
	      /* print usage and quit */

	      s = catgets (cat, 0, 0, "Display the contents of a text file one screen at a time");
	      printf ("MORE: %s\n", s);
	      usage (cat);
	      catclose (cat);
	      exit (0);
	      break;

	    default:
	      /* Not a recognized option */

	      s = catgets (cat, 1, 0, "Not a recognized option");
	      printf ("MORE: %s: %s\n", argv[i], s);
	      usage (cat);
	      catclose (cat);
	      exit (1);
	      break;
	    } /* switch */
	} /* if argv */

      /* Assume a file.  Open it and display it. */

      /* Note that findfirst, findnext only return the base file
         name in ff_name, and not the path.  The search still
         works with a path, but you don't get it back.  So we
         have to preserve the full path to the file using
         _splitpath and _makepath. */

      _splitpath (argv[i], drive, dir, file, ext);
      done = findfirst (argv[i], &ffblk, 0);

      if (done)
	{
	  /* We were not able to find a file. Display a message and
	     set the exit value. */

	  s = catgets (cat, 1, 1, "No such file");
	  fprintf (stderr, "MORE: %s: %s\n", argv[i], s);
	  exitval = 1;
	}

      while (!done)
	{
	  /* We have found a file, so open it and display it.  Set
	     the exit value to 'successful' */

	  _makepath (fullpath, drive, dir, ffblk.ff_name, "");
	  pfile = fopen (fullpath, "r");

	  if (pfile)
	    {
	      next = more (pfile, ffblk.ff_name, prompt);
	      fclose (pfile);
	      exitval = 0;

#if 0 /* debugging */
	      printf ("** next=%d\n", next);
	      if (next)
		{
		  printf ("** we have next file\n");
		}
	      else
		{
		  printf ("** we are done\n");
		}
	      getch();
#endif
	    }

	  else
	    {
	      s = catgets (cat, 1, 2, "Cannot open file");
	      fprintf (stderr, "MORE: %s: %s\n", ffblk.ff_name, s);
	      exitval = 1;
	    }

	  /* Find next file */
	  /* but only if user did not press 'Q' during more() */

	  if (next != 0)
	    {
	      done = findnext (&ffblk);
	    }
	  else
	    {
	      done = -1;
	      i = argc;			/* allows break out of for() */
	    }
	} /* while */
    } /* for */

  /* Done */

  catclose (cat);
  exit (exitval);
  return (0);				/* keeps compiler happy */
}

/* A function to display the contents of a text file, one screenful
   at a time. */

/* Returns: 0=quit program immediately, 1=next file
   -1(reserved)=previous file (not currently used) */

int
more (FILE *pfile, const char *descr, const char *prompt)
{
  int ch;				/* the character read from file    */
  int key;				/* the key pressed by user         */
  int nchars = 0;			/* no. of chars printed per line   */
  int nlines = 0;			/* no. of lines printed per screen */

  while ((ch = fgetc (pfile)) != EOF)
    {
      putchar (ch);
      nchars++;

      /* Determine if we have caused a new line to be printed.
         Printing a CR/LF will do it, as will a screen wrap. */

      if ((ch == '\n') || (nchars == COLS))
	{
	  nlines++;
	  nchars = 0;

	  if (nlines == (LINES-1))
	    {
	      /* We have met the printable screen boundary.  Display
	         a prompt, and wait for a keypress before we show more. */

	      /* Since we don't necessarily know if the user is
	         redirecting to a file, always display on stderr. */

	      fprintf (stderr, "-- %s -- %s --", prompt, descr);
	      key = keypress();
	      putchar ('\n');
	      nlines = 0;

#if 0 /* debugging */
	      fprintf (stderr, "key=%d,%c\n", key, key);
#endif /* debugging */

	      /* Do we quit or move to next file? */

	      switch (key)
		{
		case CTRL_C:
		  fprintf (stderr, "^C");
		case 'q':
		case 'Q':
		  return (0);

		case 'n':
		case 'N':
		  return (1);
		} /* switch */
	    } /* if NLINES */
	} /* if COLS */
    } /* while */

  /* Since we can show more than one file at a time, we need to
     display a prompt when we are done with a file. */

  fprintf (stderr, "-- %s --", descr);
  key = keypress();
  putchar ('\n');

  /* Do we quit or move to next file? */

  switch (key)
    {
    case CTRL_C:
      fprintf (stderr, "^C");
    case 'q':
    case 'Q':
      return (0);

    case 'n':
    case 'N':
      return (1);
    } /* switch */
}

/* Retrieve the next keypress */

unsigned
keypress (void)
{
#if 1 /* Use BIOS services */
  unsigned key;

  while (! _bios_keybrd(_KEYBRD_READY)) /* wait */ ;
  key = _bios_keybrd(_KEYBRD_READ);
  return (key & 0xFF);
#endif

#if 0 /* Use CONIO */
  return (getch());
#endif 0
}

/* A function to display the program's usage */

void
usage (nl_catd cat)
{
  char *s;

  /* Show version, copyright, and GNU GPL */

  printf ("MORE 3.4, Copyright (C) 1994-2002 Jim Hall <jhall@freedos.org>\n");
  printf ("  This is free software, and you are welcome to redistribute it\n");
  printf ("  under the GNU GPL; see the COPYING file for details.\n");

  /* Show usage */

  s = catgets (cat, 0, 1, "Usage");
  printf ("\n%s:\n", s);

  s = catgets (cat, 0, 2, "command");
  printf ("  %s | MORE\n", s);

  s = catgets (cat, 0, 3, "file");
  printf ("  MORE %s..\n", s);
  printf ("  MORE < %s\n", s);

  /* Show available keys, while viewing a file */

  s = catgets (cat, 0, 4, "Available keys");
  printf ("\n%s:\n", s);

  s = catgets (cat, 0, 9, "Space");
  printf ("  %s = ", s);
  s = catgets (cat, 0, 10, "Next page");
  printf ("%s\n", s);

  s = catgets (cat, 0, 6, "Next file");
  printf ("  N n = %s\n", s);

  s = catgets (cat, 0, 8, "Quit program");
  printf ("  Q q = %s\n", s);
}
